#include <avr/io.h>
#include "clock.h"
#include "interface.h"
#include "mfx.h"
#include "monitor.h"
#include "motorola.h"
#include "output.h"
#include "serial.h"
#include "s88.h"

static uint8_t cmd_buffer[15];
static uint8_t cmd_length = 0;
static uint8_t cmd_read_pos = 0;
static uint16_t baud_rate = 9600;
static uint16_t baud_change = 0;
static uint32_t baud_changed_at = 0;

static uint8_t dispatch_command(const uint8_t *, uint8_t);
uint8_t interface_command(const uint8_t *, uint8_t);

void interface_init(void)
{
	DDRB |= 0x01;
	DDRD = (DDRD&0xFC)|0x02;

	serial_init(baud_rate);
}

void interface_check(void)
{
	if(serial_read_overrun())
		interface_send1(RECEIVE_OVERRUN);

	uint8_t count = serial_read_available();
	if(count>0)
	{
		PORTB |= 0x01;
		if(cmd_length==0)
		{
			uint8_t l = ~serial_read();
			if(l==0)
				serial_write(0xFF);
			else if(l>=0x10)
				interface_send1(FRAMING_ERROR);
			else
			{
				cmd_length = l;
				--count;
				cmd_read_pos = 0;
			}
		}

		if(cmd_read_pos<cmd_length)
		{
			if(cmd_read_pos+count>cmd_length)
				count = cmd_length-cmd_read_pos;
			for(uint8_t i=0; i<count; ++i)
				cmd_buffer[cmd_read_pos++] = serial_read();

			if(cmd_read_pos>=cmd_length)
			{
				uint8_t result = dispatch_command(cmd_buffer, cmd_length);
				interface_send1(result);
				cmd_length = 0;
			}
		}
		PORTB &= ~0x01;
	}

	if(baud_change)
	{
		if(!baud_changed_at)
		{
			baud_changed_at = clock_get();
			serial_set_baud(baud_change);
		}
		else if(clock_get()-baud_changed_at>3*CLOCK_RATE)
		{
			baud_change = 0;
			serial_set_baud(baud_rate);
			interface_send1(BAUD_CHANGE_FAILED);
		}
	}
}

static uint8_t dispatch_command(const uint8_t *cmd, uint8_t length)
{
	uint8_t type = cmd[0]>>4;
	if(type==0)
	{
		uint8_t subtype = (cmd[0]>>3)&1;
		if(subtype==0)
			return output_command(cmd, length);
		else
			return monitor_command(cmd, length);
	}
	else if(type==1)
		return motorola_command(cmd, length);
	else if(type==2)
		return mfx_command(cmd, length);
	else if(type==3)
		return s88_command(cmd, length);
	else if(type==7)
		return interface_command(cmd, length);
	else
		return INVALID_COMMAND;
}

void interface_send(const uint8_t *cmd, uint8_t length)
{
	serial_write(~length);
	for(uint8_t i=0; i<length; ++i)
		serial_write(cmd[i]);
}

void interface_send1(uint8_t cmd)
{
	serial_write(0xFE);
	serial_write(cmd);
}

uint8_t interface_command(const uint8_t *cmd, uint8_t length)
{
	if(cmd[0]==SET_BAUD_RATE)
	{
		if(length!=3)
			return LENGTH_ERROR;

		uint16_t baud = (cmd[1]<<8)|cmd[2];
		if(baud!=baud_rate)
		{
			if(baud!=baud_change)
			{
				baud_change = baud;
				baud_changed_at = 0;
			}
			else
			{
				baud_rate = baud_change;
				baud_change = 0;
			}
		}
	}
	else
		return INVALID_COMMAND;

	return COMMAND_OK;
}
